//============================================================================
// Name        : Bds.cpp
// Author      : Zdeněk Kalčík
// Copyright   : @2009@
// Project 		 : BDS
//============================================================================

#include <sstream>
#include <iostream>
#include <stdlib.h>
#include <vector>
using namespace std;

#include "cudd.h"
#include "FileUtil.hpp"
#include "Decompositions.hpp"
#include "Decomp.hpp"
#include "Manager.hpp"
#include "Manager2.hpp"

void printProgramParams();

int main(int argc, char* argv[]) {
	int decoposedFilePosition = 0;
	int filePosition = 0;
	int saveBlifPosition = 0;
	int saveDotPosition = 0;
	int algPosition = 0;
	char reorderAlgoritm;

	for (int i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'd': {
					if (argc <= i + 1) {
						//missing filename for decomposed function
						printProgramParams();
					} else {
						decoposedFilePosition = i + 1;
						break;
					}
				}
				case 'f': {
					if (argc <= i + 1) {
						//missing file name
						printProgramParams();
					} else {
						filePosition = i + 1;
						break;
					}
				}
				case 'i': {
					if (argc <= i + 1) {
						//missing blif save file name
						printProgramParams();
					} else {
						saveBlifPosition = i + 1;
						break;
					}
				}
				case 'o': {
					if (argc <= i + 1) {
						//missing dot save file name
						printProgramParams();
					} else {
						saveDotPosition = i + 1;
						break;
					}
				}
				case 'a': {
					if (argc <= i + 1) {
						//missing algorith number
						printProgramParams();
					} else {
						algPosition = i + 1;
						break;
					}
				}
				default: {
					printProgramParams();
				}
			}
		}
	}

	if (filePosition == 0) {
		printProgramParams();
	}

	if (algPosition != 0) {
		reorderAlgoritm = argv[algPosition][0];
	} else {
		//Default reorder algoritm
		reorderAlgoritm = '3';
	}

	cout << "\nBDD decomposition by Zdenek Kalcik @2009" << endl;
	cout << "Cudd version: ";
	Cudd_PrintVersion(stdout);
	cout << endl;

	//inicialization manager
	DdManager *manager = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, CUDD_CACHE_SLOTS, 0);

	init(manager);

	//read file
	fileStructureNew structure = readStoreFile(argv[filePosition]);

	//build BDD in main manager
	structure = buildBDD(structure);

	if (decoposedFilePosition != 0) {
		initDecompositionManager(structure.inputNames);
		FILE *file = fopen("bds.temp", "w");
		writeHeadNew(structure, file);

		for (unsigned int i = 0; i < structure.outputFunctions.size(); i++) {
			DdManager *newManager = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, 600, 0);
			DdNode *function = copyFunction(structure.outputFunctions[i], newManager, true);
			if (Cudd_IsComplement(structure.outputFunctions[i])) {
				DdNode *tmp = Cudd_Complement(function);
				Cudd_Ref(tmp);
				Cudd_RecursiveDeref(newManager, function);
				function = tmp;
			}

			Decomp2 *decomp = new Decomp2();
			decompositionManager(newManager, function, decomp);
			proceedDecompp(decomp, file, structure.outputNames[i]);
			delete decomp;
		}

		writeEnd(structure, argv[decoposedFilePosition], file);

		endDecompositionManager();
		/*//copy output functions into Decomp with it's own manager (original functions perservere in main manager)
		 vector<Decomp*> decomps = copyFunctions(structure.outputFunctions, manager);

		 FILE *file = fopen("bds.temp", "w");
		 writeHead(structure, file);
		 for (unsigned int i = 0; i < decomps.size(); i++) {
		 doDecomposition(decomps[i], reorderAlgoritm);
		 //cout << endl << structure.outputNames[i] << " = ";
		 proceedDecomp(decomps[i], file, structure.outputNames[i]);
		 delete decomps[i];
		 cout << endl;
		 }
		 writeEnd(structure, argv[decoposedFilePosition], file);
		 //start decomposition, which create factoring trees from Decomp in decomps
		 //doDecomposition(decomps, reorderAlgoritm);

		 //writeDecomp(decomps, argv[decoposedFilePosition], structure);

		 cout << "blif from decomponed function writed to " << argv[decoposedFilePosition] << " successfuly" << endl;*/
	}

	//call reorder on the main BDD
	//cout << endl << "Statistics before reordering:";
	//printStatistics(manager);
	reorder(manager, reorderAlgoritm);
	//cout << endl << "Statistics after reordering:";
	//printStatistics(manager);

	//save main BDD to blif file (if wanted)
	if (saveBlifPosition != 0) {
		writeBliff(manager, structure, argv[saveBlifPosition]);
		cout << "blif from BDD writed to " << argv[saveBlifPosition] << " successfuly" << endl;
	}

	//save main BDD to dot file (if wanted)
	if (saveDotPosition != 0) {
		writeDot(manager, structure, argv[saveDotPosition]);
		cout << "dot from BDD writed to " << argv[saveDotPosition] << " successfuly" << endl;
	}

	cout << "\n\n*********END************" << endl;

	return 0;
}

/**
 * Print program parametres.
 */
void printProgramParams() {
	cerr << "Program parameters: " << endl;
	cerr << "  -f <required> filename (must be in blif format)" << endl;
	cerr << "  -d <optional> filename for saving decomposed function in blif " << endl;
	cerr << "  -i <optional> filename for saving in blif" << endl;
	cerr << "  -o <optional> filename for saving in dot" << endl;
	cerr << "  -a <optional> type of reordering algorithm (default is CUDD_REORDER_SIFT)" << endl;
	cerr << "       1 = CUDD_REORDER_SAME" << endl;
	cerr << "       2 = CUDD_REORDER_RANDOM" << endl;
	cerr << "       3 = CUDD_REORDER_SIFT" << endl;
	cerr << "       4 = CUDD_REORDER_SYMM_SIFT" << endl;
	cerr << "       5 = CUDD_REORDER_GROUP_SIFT" << endl;
	cerr << "       6 = CUDD_REORDER_ANNEALING" << endl;
	cerr << "       7 = CUDD_REORDER_GENETIC" << endl;
	cerr << "       8 = CUDD_REORDER_LINEAR" << endl;
	cerr << "       9 = CUDD_REORDER_EXACT" << endl;

	exit(1);
}
