//============================================================================
// Name        : Decompositions.cpp
// Author      : Zdeněk Kalčík
// Copyright   : @2009@
// Project 		 : BDS
//============================================================================

#include "Decompositions.hpp"

//VARIABLES
//set for variable index above cut
set<int> cut;
map<string, DdNode*> tempMap;
map<string, DdNode*>::iterator iteratorStringNode;

//pointer to generalized x-dominator
DdNode *genxdom;

int managerCount = 1;

char reorderType;

//PRIVATE METHODS
DdNode * rekAnd(DdNode *node, DdManager *managerVedlejsi, bool first);
DdNode * rekOr(DdNode *node, DdManager *managerVedlejsi, bool first);
void findGenXDom(DdNode *node);
bool isUnderCut(DdNode *node);

void initCut(set<int> cut2) {
	cut = cut2;
}

void clearCut() {
	cut.clear();
}

void addToCut(int index) {
	cut.insert(index);
}

void initGNX(DdNode *node) {
	genxdom = node;
}

void initReorderType(char type) {
	reorderType = type;
}

functions newAND(DdManager *originalManager, DdNode *originalFunction, DdManager *newManager) {
	//convert na ADD
	DdNode *tmp = Cudd_BddToAdd(originalManager, originalFunction);
	Cudd_Ref(tmp);
	Cudd_RecursiveDeref(originalManager, originalFunction);

	tempMap.clear();
	DdNode* newFunction = rekAnd(tmp, newManager, true);
	Cudd_ReduceHeap(originalManager, CUDD_REORDER_SIFT, 1);

	//convert to BDD
	DdNode *temp = Cudd_addBddPattern(originalManager, tmp);
	Cudd_Ref(temp);
	Cudd_RecursiveDeref(originalManager, tmp);

	functions fs;
	fs.newFunction = newFunction;
	fs.originalFunction = temp;

	return fs;
}

functions newOR(DdManager *originalManager, DdNode *originalFunction, DdManager *newManager) {
	//convert na ADD
	DdNode *tmp = Cudd_BddToAdd(originalManager, originalFunction);
	Cudd_Ref(tmp);
	Cudd_RecursiveDeref(originalManager, originalFunction);

	tempMap.clear();
	DdNode* newFunction = rekOr(tmp, newManager, true);
	Cudd_ReduceHeap(originalManager, CUDD_REORDER_SIFT, 1);

	//convert to BDD
	DdNode *temp = Cudd_addBddPattern(originalManager, tmp);
	Cudd_Ref(temp);
	Cudd_RecursiveDeref(originalManager, tmp);

	functions fs;
	fs.newFunction = newFunction;
	fs.originalFunction = temp;

	return fs;
}

functions newXNOR(DdManager *originalManager, DdNode *originalFunction, DdManager *newManager) {

	tempMap.clear();
	DdNode* newFunction = copyFunction(genxdom, newManager, true);

	//create function H = G XNOR F
	DdNode *xnor = Cudd_bddXnor(originalManager, originalFunction, genxdom);
	Cudd_Ref(xnor);
	Cudd_RecursiveDeref(originalManager, originalFunction);
	Cudd_ReduceHeap(originalManager, CUDD_REORDER_SIFT, 1);

	functions fs;
	fs.newFunction = newFunction;
	fs.originalFunction = xnor;

	return fs;
}

/**
 * This method makes AND decomposition.
 * It requires set<int> cut to be filled (index of variables obove the cut).
 */
void doAND(Decomp *decomp) {
	DdManager *managerVedlejsi = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, MANAG_CACHE_SIZE, 0);
	//cout << "manager count: " << ++managerCount << endl;

	//convert na ADD
	DdNode *tmp = Cudd_BddToAdd(decomp -> manag, decomp -> function);
	Cudd_Ref(tmp);
	Cudd_IterDerefBdd(decomp -> manag, decomp -> function);
	decomp -> function = tmp;

	tempMap.clear();
	DdNode* copy = rekAnd(decomp -> function, managerVedlejsi, true);
	reorder(decomp -> manag, reorderType);
	//reorder(managerVedlejsi, reorderType);

	//convert to BDD
	tmp = Cudd_addBddPattern(decomp -> manag, decomp -> function);
	Cudd_Ref(tmp);
	Cudd_IterDerefBdd(decomp -> manag, decomp -> function);
	decomp -> function = tmp;

	//store function Q and D
	Decomp *decompL = new Decomp(decomp);
	Decomp *decompR = new Decomp(copy, managerVedlejsi);
	decomp -> saveDecomp(decompL, decompR, AND);

	//writeDotForDecomp(decompL, "Q.dot");
	//writeDotForDecomp(decompR, "D.dot");
}

/**
 * This method makes OR decomposition.
 * It requires set<int> cut to be filled (index of variables above the cut).
 */
void doOR(Decomp *decomp) {
	DdManager *managerVedlejsi = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, MANAG_CACHE_SIZE, 0);
	//cout << "manager count: " << ++managerCount << endl;

	//convert na ADD
	DdNode *tmp = Cudd_BddToAdd(decomp -> manag, decomp -> function);
	Cudd_Ref(tmp);
	Cudd_IterDerefBdd(decomp -> manag, decomp -> function);
	decomp -> function = tmp;

	tempMap.clear();
	DdNode* copy = rekOr(decomp -> function, managerVedlejsi, true);
	reorder(decomp -> manag, reorderType);
	//reorder(managerVedlejsi, reorderType);

	//convert to BDD
	tmp = Cudd_addBddPattern(decomp -> manag, decomp -> function);
	Cudd_Ref(tmp);
	Cudd_IterDerefBdd(decomp -> manag, decomp -> function);
	decomp -> function = tmp;

	//store function G and H
	Decomp *decompL = new Decomp(decomp);
	Decomp *decompR = new Decomp(copy, managerVedlejsi);
	decomp -> saveDecomp(decompL, decompR, OR);

	//writeDotForDecomp(decompL, "H.dot");
	//writeDotForDecomp(decompR, "G.dot");
}

/**
 * This method makes XNOR decomposition.
 * It requires string gnx to be filled (name of generalized x-dominator).
 */
void doXNOR(Decomp *decomp) {
	//create function G with new manager
	DdManager *managerVedlejsi = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, MANAG_CACHE_SIZE, 0);
	//cout << "manager count: " << ++managerCount << endl;

	tempMap.clear();
	DdNode* copy = copyFunction(genxdom, managerVedlejsi, true);

	//create function H = G XNOR F
	DdNode *xnor = Cudd_bddXnor(decomp -> manag, decomp -> function, genxdom);
	Cudd_Ref(xnor);
	Cudd_RecursiveDeref(decomp -> manag, decomp -> function);
	decomp -> function = xnor;
	reorder(decomp -> manag, reorderType);

	//store function G and H
	Decomp *decompL = new Decomp(decomp);
	Decomp *decompR = new Decomp(copy, managerVedlejsi);
	decomp -> saveDecomp(decompL, decompR, XNOR);

	//writeDotForDecomp(decompL, "H.dot");
	//writeDotForDecomp(decompR, "G.dot");
}

DdNode * rekAnd(DdNode *node, DdManager *managerVedlejsi, bool first) {
	DdNode *T;
	DdNode *E;
	DdNode *N;
	DdNode *F;
	DdNode *tmp;

	if (Cudd_IsConstant(Cudd_T(node))) {
		if (Cudd_V(Cudd_T(node)) == 1) {
			T = Cudd_ReadOne(managerVedlejsi);
		} else {
			tmp = Cudd_T(node);
			Cudd_T(node) = Cudd_E(node);
			Cudd_Ref(Cudd_T(node));
			Cudd_Deref(tmp);
			T = Cudd_ReadLogicZero(managerVedlejsi);
		}
	} else {
		if (isUnderCut(Cudd_T(node))) {
			T = Cudd_ReadOne(managerVedlejsi);
		} else {
			//cout << "searching: " << getNodeIndex(Cudd_T(node)) << endl;
			if (tempMap.find(getNodeIndex(Cudd_T(node))) != tempMap.end()) {
				//cout << "founded" << endl;
				T = tempMap[getNodeIndex(Cudd_T(node))];
			} else {
				T = rekAnd(Cudd_T(node), managerVedlejsi, false);
			}
		}
	}

	if (Cudd_IsConstant(Cudd_E(node))) {
		if (Cudd_V(Cudd_E(node)) == 1) {
			E = Cudd_ReadOne(managerVedlejsi);
		} else {
			tmp = Cudd_E(node);
			Cudd_E(node) = Cudd_T(node);
			Cudd_Ref(Cudd_E(node));
			Cudd_Deref(tmp);
			E = Cudd_ReadLogicZero(managerVedlejsi);
		}
	} else {
		if (isUnderCut(Cudd_E(node))) {
			E = Cudd_ReadOne(managerVedlejsi);
		} else {
			//cout << "searching: " << getNodeIndex(Cudd_E(node)) << endl;
			if (tempMap.find(getNodeIndex(Cudd_E(node))) != tempMap.end()) {
				//cout << "founded" << endl;
				E = tempMap[getNodeIndex(Cudd_E(node))];
			} else {
				E = rekAnd(Cudd_E(node), managerVedlejsi, false);
			}
		}
	}

	N = Cudd_bddIthVar(managerVedlejsi, node -> index);
	F = Cudd_bddIte(managerVedlejsi, N, T, E);
	Cudd_Ref(F);

	if (first) {
		for (iteratorStringNode = tempMap.begin(); iteratorStringNode != tempMap.end(); iteratorStringNode++) {
			Cudd_IterDerefBdd(managerVedlejsi, (*iteratorStringNode).second);
		}
	} else {
		tempMap[getNodeIndex(node)] = F;
	}

	//tempMap[getNodeIndex(node)] = F;
	//cout << "done: " << getNodeIndex(node) << endl;
	//Cudd_IterDerefBdd(managerVedlejsi, T);
	//Cudd_IterDerefBdd(managerVedlejsi, E);
	return F;
}

DdNode * rekOr(DdNode *node, DdManager *managerVedlejsi, bool first) {
	DdNode *T;
	DdNode *E;
	DdNode *N;
	DdNode *F;
	DdNode *tmp;

	if (Cudd_IsConstant(Cudd_T(node))) {
		if (Cudd_V(Cudd_T(node)) == 1) {
			tmp = Cudd_T(node);
			Cudd_T(node) = Cudd_E(node);
			Cudd_Ref(Cudd_T(node));
			Cudd_Deref(tmp);
			T = Cudd_ReadOne(managerVedlejsi);
		} else {
			T = Cudd_ReadLogicZero(managerVedlejsi);
		}
	} else {
		if (isUnderCut(Cudd_T(node))) {
			T = Cudd_ReadLogicZero(managerVedlejsi);
		} else {
			if (tempMap.find(getNodeIndex(Cudd_T(node))) != tempMap.end()) {
				T = tempMap[getNodeIndex(Cudd_T(node))];
			} else {
				T = rekOr(Cudd_T(node), managerVedlejsi, false);
			}
		}
	}

	if (Cudd_IsConstant(Cudd_E(node))) {
		if (Cudd_V(Cudd_E(node)) == 1) {
			tmp = Cudd_E(node);
			Cudd_E(node) = Cudd_T(node);
			Cudd_Ref(Cudd_E(node));
			Cudd_Deref(tmp);
			E = Cudd_ReadOne(managerVedlejsi);
		} else {
			E = Cudd_ReadLogicZero(managerVedlejsi);
		}
	} else {
		if (isUnderCut(Cudd_E(node))) {
			E = Cudd_ReadLogicZero(managerVedlejsi);
		} else {
			if (tempMap.find(getNodeIndex(Cudd_E(node))) != tempMap.end()) {
				E = tempMap[getNodeIndex(Cudd_E(node))];
			} else {
				E = rekOr(Cudd_E(node), managerVedlejsi, false);
			}
		}
	}

	N = Cudd_bddIthVar(managerVedlejsi, node -> index);
	F = Cudd_bddIte(managerVedlejsi, N, T, E);
	Cudd_Ref(F);
	if (first) {
		for (iteratorStringNode = tempMap.begin(); iteratorStringNode != tempMap.end(); iteratorStringNode++) {
			Cudd_IterDerefBdd(managerVedlejsi, (*iteratorStringNode).second);
		}
	} else {
		tempMap[getNodeIndex(node)] = F;
	}
	//tempMap[getNodeIndex(node)] = F;
	//Cudd_IterDerefBdd(managerVedlejsi, T);
	//Cudd_IterDerefBdd(managerVedlejsi, E);
	return F;
}

bool isUnderCut(DdNode *node) {
	if (cut.find(node -> index) == cut.end()) {
		return true;
	} else {
		return false;
	}
}

/**
 * Create new Decomp for every function with it's own manager.
 */
vector<Decomp*> copyFunctions(vector<DdNode*> functions, DdManager *managerOld) {
	vector<Decomp*> decomps;
	Decomp * decomp;
	DdNode *copy, *tmp;
	DdManager *managerNew;
	for (unsigned int i = 0; i < functions.size(); i++) {
		tempMap.clear();
		managerNew = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, MANAG_CACHE_SIZE, 0);
		copy = copyFunction(functions[i], managerNew, true);
		if (Cudd_IsComplement(functions[i])) {
			tmp = Cudd_Complement(copy);
			Cudd_Ref(tmp);
			Cudd_RecursiveDeref(managerNew, copy);
			copy = tmp;
		}
		decomp = new Decomp(copy, managerNew);
		decomps.push_back(decomp);
	}
	return decomps;
}

/**
 * Returns unique node index.
 */
string getNodeIndex(DdNode *node) {
	stringstream out;
	out << node;
	return out.str();
}

DdNode * copyFunction(DdNode *node, DdManager *managerNew, bool first) {
	DdNode *T;
	DdNode *E;
	DdNode *N;
	DdNode *F;

	if (Cudd_IsConstant(Cudd_T(node))) {
		if (Cudd_V(Cudd_T(node)) == 1) {
			T = Cudd_ReadOne(managerNew);
		} else {
			T = Cudd_ReadLogicZero(managerNew);
		}
	} else {
		//cout << "searching: " << getNodeIndex(Cudd_T(node)) << endl;
		if (tempMap.find(getNodeIndex(Cudd_T(node))) != tempMap.end()) {
			//cout << "founded" << endl;
			T = tempMap[getNodeIndex(Cudd_T(node))];
		} else {
			T = copyFunction(Cudd_T(node), managerNew, false);
		}
	}

	if (Cudd_IsConstant(Cudd_E(node))) {
		if (Cudd_V(Cudd_E(node)) == 1) {
			E = Cudd_ReadOne(managerNew);
		} else {
			E = Cudd_ReadLogicZero(managerNew);

		}
	} else {

		if (Cudd_IsComplement(Cudd_E(node))) {
			//cout << "searching: " << getNodeIndex(Cudd_Regular(Cudd_E(node))) << endl;
			if (tempMap.find(getNodeIndex(Cudd_Regular(Cudd_E(node)))) != tempMap.end()) {
				E = tempMap[getNodeIndex(Cudd_Regular(Cudd_E(node)))];
			} else {
				E = copyFunction(Cudd_Regular(Cudd_E(node)), managerNew, false);
			}
		} else {
			//cout << "searching: " << getNodeIndex(Cudd_E(node)) << endl;
			if (tempMap.find(getNodeIndex(Cudd_E(node))) != tempMap.end()) {
				E = tempMap[getNodeIndex(Cudd_E(node))];
			} else {
				E = copyFunction(Cudd_E(node), managerNew, false);
			}
		}
	}

	if (Cudd_IsComplement(Cudd_E(node))) {
		E = Cudd_Complement(E);
	}

	if (Cudd_IsComplement(node)) {
		N = Cudd_bddIthVar(managerNew, Cudd_Regular(node) -> index);
	} else {
		N = Cudd_bddIthVar(managerNew, node -> index);
	}
	F = Cudd_bddIte(managerNew, N, T, E);
	Cudd_Ref(F);
	//cout << "done: " << getNodeIndex(node) << endl;
	if (first) {
		for (iteratorStringNode = tempMap.begin(); iteratorStringNode != tempMap.end(); iteratorStringNode++) {
			Cudd_IterDerefBdd(managerNew, (*iteratorStringNode).second);
		}
	} else {
		tempMap[getNodeIndex(node)] = F;
	}
	//Cudd_IterDerefBdd(managerNew, T);
	//Cudd_IterDerefBdd(managerNew, E);
	//Cudd_Deref(T);
	//Cudd_Deref(E);
	return F;
}

/**
 * Reorder procedure.
 */
void reorder(DdManager* manager, char reorderAlg) {
	//cout << "reordering with: ";
	switch (reorderAlg) {
		case '1': {
			//cout << "CUDD_REORDER_SAME" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_SAME, 1);
			break;
		}
		case '2': {
			//cout << "CUDD_REORDER_RANDOM" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_RANDOM, 1);
			break;
		}
		case '3': {
			//cout << "CUDD_REORDER_SIFT" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_SIFT, 1);
			break;
		}
		case '4': {
			//cout << "CUDD_REORDER_SYMM_SIFT" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_SYMM_SIFT, 1);
			break;
		}
		case '5': {
			//cout << "CUDD_REORDER_GROUP_SIFT" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_GROUP_SIFT, 1);
			break;
		}
		case '6': {
			//cout << "CUDD_REORDER_ANNEALING" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_ANNEALING, 1);
			break;
		}
		case '7': {
			//cout << "CUDD_REORDER_GENETIC" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_GENETIC, 1);
			break;
		}
		case '8': {
			//cout << "CUDD_REORDER_LINEAR" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_LINEAR, 1);
			break;
		}
		case '9': {
			//cout << "CUDD_REORDER_EXACT" << endl;
			Cudd_ReduceHeap(manager, CUDD_REORDER_EXACT, 1);
			break;
		}

		default: {
			cerr << "Bad type of reordering algorithm" << endl;
			exit(1);
		}
	}
}

