//============================================================================
// Name        : FileUtil.cpp
// Author      : Zdeněk Kalčík
// Copyright   : @2009@
// Project 		 : BDS
//============================================================================

#include "FileUtil.hpp"
#include "Entry.hpp"
#include "TextUtil.hpp"

//VARIABLES
bool enableDebug = true;
DdManager* mainManager;
bool convertToADD = false;

char** inputNames;
int countAND = 0;
int countOR = 0;
int countXNOR = 0;

int managerDestroyed = 0;

//PRIVATE METHODS
void readVariablesNames(string line, vector<string> & namesVector);
void printDebug(string text);
void printDebugLn(string text);
DdNode* getOutNode(string name);
DdNode* getInNode(string name);
void saveNode(string name, DdNode* node);
void addOutput(string name, map<string, int> &variableCount);
//void proceedDecomp(Decomp *decomp, FILE *file, string funcName);
void writeBodyDecomp(Decomp *decomp, FILE *file, string funcName);
void findAndReplaceAllConstants(fileStructureNew structure, char* fileName);

//MAPS
map<string, int> inputNodes;
map<string, DdNode *> outputNodes;
map<string, DdNode *> innerNodes;

set<string> inicializedNodes;

string con0, con1;

/**
 * Print statistics for manager.
 */
void printStatistics(DdManager *manag) {
	if (enableDebug) {
		cout << "\n-----------------------------------------------------------\n";
		cout << "num variables in existence: " << Cudd_ReadSize(manag) << "\n";
		cout << "node count: " << Cudd_ReadNodeCount(manag) << "\n";
		cout << "num reorderings: " << Cudd_ReadReorderings(manag) << "\n";
		cout << "memory in use: " << Cudd_ReadMemoryInUse(manag) << " bytes \n";
		cout << "-----------------------------------------------------------\n\n";
	}
}

void init(DdManager* manager1) {
	mainManager = manager1;
}

/**
 * Parse line and store names in namesVector.
 */
void readVariablesNames(string line, vector<string> & namesVector) {
	size_t found;
	found = line.find_first_of(' ');
	string temp;
	while (found != string::npos) {
		temp = line.substr(0, found);
		namesVector.push_back(temp);
		line = line.substr(found + 1);
		line = trimS(line);
		found = line.find_first_of(' ');
	}
	//last substring is last name of variable
	namesVector.push_back(line.substr(0, line.length()));
}

void writeDot(DdManager *manager, fileStructureNew structure, char* fileName) {
	FILE *file = fopen(fileName, "w");

	DdNode **outputs = new DdNode*[structure.outputFunctions.size()];
	for (unsigned int i = 0; i < structure.outputFunctions.size(); i++) {
		outputs[i] = structure.outputFunctions[i];
	}

	char** inputNames = new char*[structure.inputNames.size()];
	for (unsigned int i = 0; i < structure.inputNames.size(); i++) {
		inputNames[i] = new char[structure.inputNames[i].size()];
		strcpy(inputNames[i], structure.inputNames[i].c_str());
	}

	char** outputNames = new char*[structure.outputNames.size()];
	for (unsigned int i = 0; i < structure.outputNames.size(); i++) {
		outputNames[i] = new char[structure.outputNames[i].size()];
		strcpy(outputNames[i], structure.outputNames[i].c_str());
	}

	Cudd_DumpDot(manager, structure.outputFunctions.size(), outputs, inputNames, outputNames, file);
}

void writeBliffOld(DdManager* manager, fileStructureNew structure, char* fileName) {
	FILE *file = fopen(fileName, "w");

	char *model = new char[structure.model.size()];
	strcpy(model, structure.model.c_str());

	DdNode **outputs = new DdNode*[structure.outputFunctions.size()];
	for (unsigned int i = 0; i < structure.outputFunctions.size(); i++) {
		outputs[i] = structure.outputFunctions[i];
	}

	char** inputNames = new char*[structure.inputNames.size()];
	for (unsigned int i = 0; i < structure.inputNames.size(); i++) {
		inputNames[i] = new char[structure.inputNames[i].size()];
		strcpy(inputNames[i], structure.inputNames[i].c_str());
	}

	char** outputNames = new char*[structure.outputNames.size()];
	for (unsigned int i = 0; i < structure.outputNames.size(); i++) {
		outputNames[i] = new char[structure.outputNames[i].size()];
		strcpy(outputNames[i], structure.outputNames[i].c_str());
	}

	Cudd_DumpBlif(manager, structure.outputFunctions.size(), outputs, inputNames, outputNames, model, file, 0);
}

void writeBliff(DdManager* manager, fileStructureNew structure, char* fileName) {

	FILE *file = fopen(fileName, "w");

	//.model
	fputs(".model ", file);
	fputs(structure.model.c_str(), file);
	fputs("\n", file);

	//.inputs
	inputNames = new char*[structure.inputNames.size()];
	fputs(".inputs", file);
	for (unsigned i = 0; i < structure.inputNames.size(); i++) {
		inputNames[i] = new char[structure.inputNames[i].size()];
		strcpy(inputNames[i], structure.inputNames[i].c_str());
		fputs(" ", file);
		fputs(structure.inputNames[i].c_str(), file);
	}
	fputs("\n", file);

	//.outputs
	char** outputNames = new char*[structure.outputNames.size()];
	fputs(".outputs", file);
	for (unsigned i = 0; i < structure.outputNames.size(); i++) {
		fputs(" ", file);
		fputs(structure.outputNames[i].c_str(), file);
		outputNames[i] = new char[structure.outputNames[i].size()];
		strcpy(outputNames[i], structure.outputNames[i].c_str());
	}
	fputs("\n", file);

	DdNode **outputs = new DdNode*[structure.outputFunctions.size()];
	for (unsigned int i = 0; i < structure.outputFunctions.size(); i++) {
		outputs[i] = structure.outputFunctions[i];
	}

	Cudd_DumpBlifBody(manager, structure.outputFunctions.size(), outputs, inputNames, outputNames, file, 0);
	fputs(".end", file);
	fclose(file);
}

void printDebug(string text) {
	if (enableDebug) {
		cout << text;
	}
}

void printDebugLn(string text) {
	if (enableDebug) {
		printDebug(text);
		cout << endl;
	}
}

/**
 * Returns output node by name, if output node doesn't exist in outputNodes then it returns inner node (if it doesn't exist in innerNodes neither
 * then this method creates new inner node, saves it in innerNodes and returns)
 */
DdNode* getOutNode(string name) {
	if (outputNodes.find(name) == outputNodes.end()) {
		//it's inner node
		//cout << "nenalezena vystupni promenna, jedna se o pomocnou" << endl;
		if (innerNodes.find(name) == innerNodes .end()) {
			//cout << "vnitrni promenna jeste neexisstuje" << endl;
			innerNodes[name] = Cudd_ReadLogicZero(mainManager);
		}
		return innerNodes[name];
	} else {
		return outputNodes[name];
	}
}

DdNode* getInNode(string name) {
	if (inputNodes.find(name) == inputNodes.end()) {
		//it's inner node
		//cout << "nenalezena vstupni promenna, jedna se o pomocnou" << endl;
		if (innerNodes.find(name) == innerNodes .end()) {
			if (outputNodes.find(name) == outputNodes.end()) {
				cerr << "Serious error in building BDD (getInNode - not initialized variable '" << name << "')" << endl;
				exit(1);
			} else {
				return outputNodes[name];
			}
		} else {
			//cout << "vnitrni promenna jiz exisstuje" << endl;
			return innerNodes[name];
		}
	} else {
		//cout << "vracim " << name << " s indexem " << inputNodes[name] << endl;
		return Cudd_bddIthVar(mainManager, inputNodes[name]);
	}
}

void saveNode(string name, DdNode* node) {
	if (outputNodes.find(name) == outputNodes .end()) {
		//it's inner node
		//cout << "save> nenalezena vystupni promenna, jedna se o pomocnou" << endl;
		if (innerNodes.find(name) == innerNodes .end()) {
			cerr << "Serious error in building BDD (saveNode - not initialized inner variable '" << name << "')" << endl;
			exit(1);
		} else {
			//cout << "save> vnitrni promenna jiz exisstuje" << endl;
			//cout << "ukladam inner: '" << name << "'" << endl;
			innerNodes[name] = node;
			//cout << "inicializovana " << name << endl;
			inicializedNodes.insert(name);
		}
	} else {
		outputNodes[name] = node;
		//cout << "inicializovana " << name << endl;
		inicializedNodes.insert(name);
	}
}

void readLine(ifstream &inFile, string & line) {
	do {
		getline(inFile, line);
		line = trimS(line);
	} while (line.length() == 0);
}

/**
 * Reads file with given fileName and creates fileStructureNew for it.
 */
fileStructureNew readStoreFile(char* fileName) {
	//open file
	ifstream inFile(fileName);
	fileStructureNew structure;

	//check, if the file exists
	if (!inFile) {
		cerr << "Can't open file: " << fileName << endl;
		exit(1);
	} else {
		printDebugLn("**************************************");
		printDebug(fileName);
		printDebugLn(" opened successful.");
	}

	//vector for savin variables names
	vector<string> variables;

	//auxiliary variables
	size_t found;
	string word, line;
	bool continueLine = true;
	bool continueCycle = true;
	bool constant = false;
	Entry* zaznam;

	vector<Entry*> nonOrderedEntries;

	map<string, int> variableCount;
	map<string, int> variablePlace;

	inFile >> word;

	while (word.compare(".model") != 0 && word.compare(".inputs") != 0) {
		inFile >> word;
		if (inFile.eof()) {
			//TODO chybi inputs
			cerr << "chybi inputs" << endl;
			exit(1);
		}
	}

	//first word should be ".model" (but it needn't)
	if (word.compare(".model") == 0) {
		inFile >> word;
		structure.model = word;
		printDebug("Model name: ");
		printDebugLn(word);
		inFile >> word;
	} else {
		printDebug("Missing model name, but it's posible to continue.");
		structure.model = "model";
	}

	//next word must be ".inputs"
	if (word.compare(".inputs") == 0) {
		printDebug("Input variable names: ");
		while (continueLine) {
			continueLine = false;
			readLine(inFile, line);
			line = trimS(line);
			variables.clear();
			readVariablesNames(line, variables);
			for (unsigned int i = 0; i < variables.size(); i++) {
				if (variables[i].compare("\\") == 0) {
					continueLine = true;
				} else {
					printDebug(variables[i]);
					printDebug(" ");
					structure.inputNames.push_back(variables[i]);
				}
			}
		}
		printDebugLn("");
		inFile >> word;
		continueLine = true;
	} else {
		//TODO chybi inputs
		cerr << "chybi inputs" << endl;
		exit(1);
	}

	//next word must be ".outputs"
	if (word.compare(".outputs") == 0) {
		printDebug("Output variable names: ");
		while (continueLine) {
			continueLine = false;
			readLine(inFile, line);
			//line = trimS(line);
			variables.clear();
			readVariablesNames(line, variables);
			//store output names in fileStructure
			for (unsigned int i = 0; i < variables.size(); i++) {
				if (variables[i].compare("\\") == 0) {
					continueLine = true;
				} else {
					printDebug(variables[i]);
					printDebug(" ");
					structure.outputNames.push_back(variables[i]);
				}
			}
		}
		printDebugLn("");
	} else {
		//TODO chybi outputs
		cerr << "chybi outputs" << endl;
		exit(1);
	}

	printDebugLn("**************************************");

	while (continueCycle) {
		readLine(inFile, line);
		line = trimS(line);
		found = line.find_first_of(' ');
		if (found == string::npos) {
			if (line.compare("0") == 0) {
				structure.con0 = variables[0];
				constant = false;
			} else if (line.compare("1") == 0) {
				structure.con1 = variables[0];
				constant = false;
			} else if (line.compare(".end") == 0) {
				printDebugLn("\nWhole file read");
				continueCycle = false;
			} else {
				//TODO spatna syntaxe
				cerr << "Spatna syntaxe" << endl;
				exit(1);
			}
		} else {
			string firstWord = line.substr(0, found);
			if (firstWord.compare(".names") == 0) {
				if (constant && variables.size() == 1) {
					structure.con0 = variables[0];
					constant = false;
				}
				variables.clear();
				line = line.substr(found + 1);
				line = trimS(line);
				readVariablesNames(line, variables);

				while (true) {
					if (variables[variables.size() - 1].compare("\\") == 0) {
						variables.pop_back();
						readLine(inFile, line);
						//line = trimS(line);
						readVariablesNames(line, variables);
					} else {
						break;
					}
				}

				if (variables.size() < 2) {
					//constant node
					constant = true;
					continue;
				}

				zaznam = new Entry(variables[variables.size() - 1]);
				addOutput(variables[variables.size() - 1], variableCount);
				for (unsigned i = 0; i < variables.size() - 1; i++) {
					zaznam->addInput(variables[i]);
				}
				nonOrderedEntries.push_back(zaznam);
			} else {
				zaznam->addSetting(line);
			}
		}
	}

	int pointPlace = 0;
	int tempCount;
	string tempName;
	vector<Entry*> orderedEntries;
	orderedEntries.reserve(nonOrderedEntries.size());

	for (unsigned int i = 0; i < nonOrderedEntries.size(); i++) {
		tempName = nonOrderedEntries[i] -> getOutput();
		if (variablePlace.find(tempName) != variablePlace.end()) {
			//vim kam muzu umistit
			tempCount = variablePlace[tempName];
			orderedEntries[tempCount] = nonOrderedEntries[i];
			variablePlace[tempName] = ++tempCount;
		} else {
			//muzu umistit na freePlace
			orderedEntries[pointPlace] = nonOrderedEntries[i];
			variablePlace[tempName] = pointPlace + 1;
			pointPlace += variableCount[tempName];
		}
	}

	for (unsigned int i = 0; i < nonOrderedEntries.size(); i++) {
		//orderedEntries[i] -> toString();
		structure.orderedEntries1.push_back(orderedEntries[i]);
	}

	structure.variableCount = variableCount;
	//cout << "pocet ruznych vystupu: " << variableCount.size() << endl;

	inFile.close();
	return structure;
}

void addOutput(string name, map<string, int> &variableCount) {
	if (variableCount.find(name) == variableCount.end()) {
		variableCount[name] = 1;
	} else {
		int count = variableCount[name];
		variableCount[name] = ++count;
	}
}

/**
 * Creates node for Entry.
 */
void buildNodeFromEntry(Entry* entry) {
	DdNode *actualOut, *cube, *var, *tmp, *tmp2, *tmp3;
	bool skipRow = false;

	actualOut = Cudd_ReadLogicZero(mainManager);
	Cudd_Ref(actualOut);
	for (int i = 0; i < entry -> lineCount(); i++) {
		string line = entry -> getLine(i);

		cube = Cudd_ReadOne(mainManager);
		Cudd_Ref(cube);
		for (unsigned int ii = 0; ii < line.length(); ii++) {

			if (con1.compare(entry -> getInput(ii)) == 0) {
				//cout << "detekovana konstanta 1, preskakuji sloupec" << endl;
				if (line.data()[ii] == '0') {
					//if it is 1-constant with output 0 then skip row
					skipRow = true;
					break;
				} else {
					//else skip column
					//cout << " preskakuji sloupec" << endl;
					continue;
				}
			}

			if (con0.compare(entry -> getInput(ii)) == 0) {
				//cout << "detekovana konstanta 0";
				if (line.data()[ii] == '1') {
					//if it is 0-constant with output 1 then skip row
					skipRow = true;
					break;
				} else {
					//else skip column
					//cout << " preskakuji sloupec" << endl;
					continue;
				}
			}

			switch (line.data()[ii]) {
				case '0': {
					var = Cudd_Not(getInNode(entry -> getInput(ii)));
					break;
				}
				case '1': {
					var = getInNode(entry -> getInput(ii));
					break;
				}
				case '-': {
					//var = Cudd_ReadOne(manager);
					continue;
				}
				default: {
					//TODO spatna syntaxe
					cerr << "Spatna syntaxe" << endl;
					exit(1);
				}
			}
			tmp = Cudd_bddAnd(mainManager, cube, var);
			Cudd_Ref(tmp);
			Cudd_IterDerefBdd(mainManager, cube);
			cube = tmp;
		}

		if (skipRow) {
			skipRow = false;
			Cudd_IterDerefBdd(mainManager, cube);
		} else {
			tmp2 = Cudd_bddOr(mainManager, actualOut, cube);
			Cudd_Ref(tmp2);
			//Cudd_RecursiveDeref(manager, actualOut);
			Cudd_IterDerefBdd(mainManager, actualOut);
			Cudd_IterDerefBdd(mainManager, cube);
			actualOut = tmp2;
		}
		//Cudd_RecursiveDeref(manager, cube);
	}

	tmp3 = Cudd_bddOr(mainManager, Cudd_NotCond(actualOut, entry -> getSingularity()), getOutNode(entry -> getOutput()));
	Cudd_Ref(tmp3);
	//Cudd_RecursiveDeref(manager, actualOut);
	//Cudd_RecursiveDeref(manager, getOutNode(entry -> getOutput()));
	Cudd_IterDerefBdd(mainManager, actualOut);
	//Cudd_IterDerefBdd(manager, getOutNode(entry -> getOutput()));
	saveNode(entry -> getOutput(), tmp3);
}

fileStructureNew buildBDD(fileStructureNew structure) {
	printDebugLn("Building BDD...");

	for (unsigned int i = 0; i < structure.inputNames.size(); i++) {
		//insert into initialized
		inicializedNodes.insert(structure.inputNames[i]);
		//insert identificator of input node
		inputNodes[structure.inputNames[i]] = i;
	}
	DdNode *node;
	for (unsigned int i = 0; i < structure.outputNames.size(); i++) {
		//initialize output
		node = Cudd_ReadLogicZero(mainManager);
		Cudd_Ref(node);
		outputNodes[structure.outputNames[i]] = node;
	}
	if (!structure.con0.empty()) {
		inicializedNodes.insert(structure.con0);
		con0 = structure.con0;
	}
	if (!structure.con1.empty()) {
		inicializedNodes.insert(structure.con1);
		con1 = structure.con1;
	}

	unsigned int entriesFinished = 0;
	for (unsigned int i = 0; i < structure.variableCount.size(); i++) {
		if (entriesFinished == structure.orderedEntries1.size()) {
			//cout << "all Entries initial" << endl;
			break;
		}
		string aktualOutput = "!@#";
		bool initNode = true;
		for (unsigned int ii = 0; ii < structure.orderedEntries1.size(); ii++) {
			if (aktualOutput.compare(structure.orderedEntries1[ii] -> getOutput()) != 0) {
				if (inicializedNodes.find(aktualOutput) != inicializedNodes.end()) {
					//cout << "promenna jiz inicializovana " << aktualOutput << endl;
				} else if (initNode && ii > 0) {
					//cout << "mohu inicializovat promennou " << aktualOutput << endl;
					for (unsigned int end = ii; end > (ii - structure.variableCount[aktualOutput]); end--) {
						buildNodeFromEntry(structure.orderedEntries1[end - 1]);
						entriesFinished++;
						//cout << "hotovo: " << entriesFinished << "/" << structure.orderedEntries1.size() << endl;
					}
				} else if (!initNode) {
					//cout << "nelze inicializovat promennou " << aktualOutput << endl;
				}
				aktualOutput = structure.orderedEntries1[ii] -> getOutput();
				//cout << "novy vystup " << aktualOutput << endl;
				initNode = true;
			} else {
				//cout << "stale kontroluji vystup " << aktualOutput << endl;
			}
			if (!structure.orderedEntries1[ii] -> canInitialize(inicializedNodes)) {
				initNode = false;
			}
		}
		if (inicializedNodes.find(aktualOutput) != inicializedNodes.end()) {
			//cout << "promenna jiz inicializovana " << aktualOutput << endl;
		} else if (initNode) {
			//cout << "posledni mohu inicializovat promennou " << aktualOutput << endl;
			//last Entry
			for (unsigned int end = structure.orderedEntries1.size(); end > (structure.orderedEntries1.size()
					- structure.variableCount[aktualOutput]); end--) {
				buildNodeFromEntry(structure.orderedEntries1[end - 1]);
				entriesFinished++;
				//cout << "hotovo: " << entriesFinished << "/" << structure.orderedEntries1.size() << endl;
			}
		} else {
			//cout << "posledni nelze inicializovat promennou " << aktualOutput << endl;
		}
	}

	for (unsigned int i = 0; i < structure.orderedEntries1.size(); i++) {
		string out = structure.orderedEntries1[i] -> getOutput();
		if (inicializedNodes.find(out) == inicializedNodes.end()) {
			cerr << "Variable " << out << " wasn't initialized." << endl;
			exit(1);
		}
	}

	for (unsigned int i = 0; i < structure.outputNames.size(); i++) {
		string out = structure.outputNames[i];
		if (inicializedNodes.find(out) == inicializedNodes.end()) {
			cerr << "Output variable " << out << " wasn't initialized." << endl;
			exit(1);
		}
	}

	DdNode *tmp;

	for (map<string, DdNode *>::iterator i = innerNodes.begin(); i != innerNodes.end(); i++) {
		Cudd_IterDerefBdd(mainManager, (*i).second);
	}

	for (unsigned int i = 0; i < structure.outputNames.size(); i++) {
		if (convertToADD) {
			//convert na ADD
			tmp = Cudd_BddToAdd(mainManager, outputNodes[structure.outputNames[i]]);
			Cudd_Ref(tmp);
			Cudd_IterDerefBdd(mainManager, outputNodes[structure.outputNames[i]]);
			structure.outputFunctions.push_back(tmp);
		} else {
			structure.outputFunctions.push_back(outputNodes[structure.outputNames[i]]);
		}
	}

	printDebugLn("BDD builded successfuly.");

	return structure;
}

void writeDecomp(vector<Decomp*> decomps, char* fileName, fileStructureNew structure) {
	FILE *file = fopen("bds.temp", "w");

	//.model
	fputs(".model ", file);
	fputs(structure.model.c_str(), file);
	fputs("\n", file);

	//.inputs
	inputNames = new char*[structure.inputNames.size()];
	fputs(".inputs", file);
	for (unsigned i = 0; i < structure.inputNames.size(); i++) {
		inputNames[i] = new char[structure.inputNames[i].size()];
		strcpy(inputNames[i], structure.inputNames[i].c_str());
		fputs(" ", file);
		fputs(structure.inputNames[i].c_str(), file);
	}
	fputs("\n", file);

	//.outputs
	char** outputNames = new char*[structure.outputNames.size()];
	fputs(".outputs", file);
	for (unsigned i = 0; i < structure.outputNames.size(); i++) {
		fputs(" ", file);
		fputs(structure.outputNames[i].c_str(), file);
		outputNames[i] = new char[structure.outputNames[i].size()];
		strcpy(outputNames[i], structure.outputNames[i].c_str());
	}
	fputs("\n", file);

	for (unsigned int i = 0; i < decomps.size(); i++) {
		cout << endl << structure.outputNames[i] << " = ";
		proceedDecomp(decomps[i], file, structure.outputNames[i]);
		cout << endl;
	}

	fputs(".end\n", file);
	fclose(file);

	findAndReplaceAllConstants(structure, fileName);
}

void writeHeadNew(fileStructureNew structure, FILE *file) {
	//.model
	fputs(".model ", file);
	fputs(structure.model.c_str(), file);
	fputs("\n", file);

	//.inputs
	fputs(".inputs", file);
	for (unsigned i = 0; i < structure.inputNames.size(); i++) {
		fputs(" ", file);
		fputs(structure.inputNames[i].c_str(), file);
	}
	fputs("\n", file);

	//.outputs
	fputs(".outputs", file);
	for (unsigned i = 0; i < structure.outputNames.size(); i++) {
		fputs(" ", file);
		fputs(structure.outputNames[i].c_str(), file);
	}
	fputs("\n", file);
}

void writeHead(fileStructureNew structure, FILE *file) {
	//.model
	fputs(".model ", file);
	fputs(structure.model.c_str(), file);
	fputs("\n", file);

	//.inputs
	inputNames = new char*[structure.inputNames.size()];
	fputs(".inputs", file);
	for (unsigned i = 0; i < structure.inputNames.size(); i++) {
		inputNames[i] = new char[structure.inputNames[i].size()];
		strcpy(inputNames[i], structure.inputNames[i].c_str());
		fputs(" ", file);
		fputs(structure.inputNames[i].c_str(), file);
	}
	fputs("\n", file);

	//.outputs
	char** outputNames = new char*[structure.outputNames.size()];
	fputs(".outputs", file);
	for (unsigned i = 0; i < structure.outputNames.size(); i++) {
		fputs(" ", file);
		fputs(structure.outputNames[i].c_str(), file);
		outputNames[i] = new char[structure.outputNames[i].size()];
		strcpy(outputNames[i], structure.outputNames[i].c_str());
	}
	fputs("\n", file);
}

void writeEnd(fileStructureNew structure, char* fileName, FILE *file) {
	fputs(".end", file);
	fclose(file);

	findAndReplaceAllConstants(structure, fileName);
}

void proceedDecomp(Decomp *decomp, FILE *file, string funcName) {
	if (decomp -> list) {
		writeBodyDecomp(decomp, file, funcName);

		//cout << "nedereferovano: " << Cudd_CheckZeroRef(decomp -> manag) << endl;
		Cudd_RecursiveDeref(decomp -> manag, decomp -> function);
		//cout << "nedereferovano: " << Cudd_CheckZeroRef(decomp -> manag) << endl;
		Cudd_Quit(decomp -> manag);

		//cout << "Zniceno manageru: " << ++managerDestroyed << endl;

	} else {
		//cout << "nalezena operace: " << decomp -> oper << endl;
		string left, right;
		switch (decomp -> oper) {
			case AND: {
				stringstream out;
				out << "Q" << countAND;
				left = out.str();
				out.str("");
				out << "D" << countAND;
				right = out.str();
				countAND++;
				fputs(".names ", file);
				fputs(left.c_str(), file);
				fputs(" ", file);
				fputs(right.c_str(), file);
				fputs(" ", file);
				fputs(funcName.c_str(), file);
				fputs("\n11 1\n", file);
				break;
			}
			case OR: {
				stringstream out;
				out << "H" << countOR;
				left = out.str();
				out.str("");
				out << "G" << countOR;
				right = out.str();
				countOR++;
				fputs(".names ", file);
				fputs(left.c_str(), file);
				fputs(" ", file);
				fputs(right.c_str(), file);
				fputs(" ", file);
				fputs(funcName.c_str(), file);
				fputs("\n1- 1\n-1 1\n", file);
				break;
			}
			case XNOR: {
				stringstream out;
				out << "XH" << countXNOR;
				left = out.str();
				out.str("");
				out << "XG" << countXNOR;
				right = out.str();
				countXNOR++;
				fputs(".names ", file);
				fputs(left.c_str(), file);
				fputs(" ", file);
				fputs(right.c_str(), file);
				fputs(" ", file);
				fputs(funcName.c_str(), file);
				fputs("\n11 1\n00 1\n", file);
				break;
			}

			default:
				cerr << "unknown opearation found" << endl;
				exit(1);
		}

		//cout << "(";
		proceedDecomp(decomp -> L, file, left);
		/*switch (decomp -> oper) {
			case AND:
				cout << "*";
				break;
			case OR:
				cout << "+";
				break;
			case XNOR:
				cout << "@";
				break;
			default:
				break;
		}*/
		proceedDecomp(decomp -> R, file, right);
		//cout << ")";
	}
}

void writeBodyDecomp(Decomp *decomp, FILE *file, string funcName) {

	if (Cudd_IsConstant(Cudd_T(decomp -> function)) && Cudd_IsConstant(Cudd_E(decomp -> function))) {

		if (Cudd_IsComplement(decomp -> function)) {
			//cout << "neg(" << inputNames[Cudd_Regular(decomp -> function) -> index] << ")";
		} else {
			//cout << inputNames[decomp -> function -> index];
		}

		fputs(".names ", file);

		if (Cudd_IsComplement(decomp -> function)) {
			fputs(inputNames[Cudd_Regular(decomp -> function) -> index], file);
		} else {
			fputs(inputNames[decomp -> function -> index], file);
		}
		fputs(" ", file);
		fputs(funcName.c_str(), file);
		if (Cudd_IsComplement(decomp -> function)) {
			fputs("\n0 1\n", file);
			//cout << " negace " << inputNames[decomp -> function -> index] << endl;
		} else {
			//cout << " normal " << inputNames[decomp -> function -> index] << endl;
			fputs("\n1 1\n", file);
		}
	} else {
		//cout << funcName;

		char** outputNames = new char*[1];
		outputNames[0] = new char[funcName.size()];
		strcpy(outputNames[0], funcName.c_str());

		DdNode **outputs = new DdNode*[1];
		outputs[0] = decomp -> function;

		Cudd_DumpBlifBody(decomp -> manag, 1, outputs, inputNames, outputNames, file, 0);
	}
}

void findAndReplaceAllConstants(fileStructureNew structure, char* fileName) {
	ifstream file("bds.temp");
	string line;
	size_t found;
	bool cycle = true;
	vector<string> temp;
	set<string> cons0;
	set<string> cons1;

	cout << endl << endl;
	readLine(file, line);
	while (cycle) {
		found = line.find(".end");
		if (found != string::npos) {
			break;
		}

		found = line.find(".names");
		if (found != string::npos) {
			temp.clear();
			readVariablesNames(line, temp);

			if (temp.size() == 2) {
				//its constant
				readLine(file, line);
				if (line.length() > 1) {
					//cout << "nalezena konstanta 0: " << temp[1] << endl;
					cons0.insert(temp[1]);
					continue;
				} else {
					if (line.compare("1") == 0) {
						//cout << "nalezena konstanta 1: " << temp[1] << endl;
						cons1.insert(temp[1]);
					} else {
						//cout << "nalezena konstanta 0: " << temp[1] << endl;
						cons0.insert(temp[1]);
					}
				}
			}

		}

		readLine(file, line);
	}

	file.close();

	if (structure.con0.empty() && cons0.size() > 0) {
		structure.con0 = "con0";
	}
	if (structure.con1.empty() && cons1.size() > 0) {
		structure.con1 = "con1";
	}

	int lineCount = 1;
	ifstream fileI("bds.temp");
	FILE *fileO = fopen(fileName, "w");
	readLine(fileI, line);
	while (cycle) {
		found = line.find(".end");
		if (found != string::npos) {
			fputs(".end\n", fileO);
			break;
		}

		found = line.find(".names");
		if (found != string::npos) {
			temp.clear();
			readVariablesNames(line, temp);

			if (temp.size() == 2) {
				//its constant -> skip
				readLine(fileI, line);
				if (line.length() == 1) {
					readLine(fileI, line);
				}
				continue;
			} else {
				fputs(".names", fileO);
				for (unsigned int i = 1; i < temp.size(); i++) {
					if (cons0.find(temp[i]) != cons0.end()) {
						fputs(" ", fileO);
						fputs(structure.con0.c_str(), fileO);
					} else if (cons1.find(temp[i]) != cons1.end()) {
						fputs(" ", fileO);
						fputs(structure.con1.c_str(), fileO);
					} else {
						fputs(" ", fileO);
						fputs(temp[i].c_str(), fileO);
					}
				}
				fputs("\n", fileO);
			}
		} else {
			fputs(line.c_str(), fileO);
			fputs("\n", fileO);

			if (lineCount == 3) {
				if (!structure.con0.empty()) {
					fputs(".names ", fileO);
					fputs(structure.con0.c_str(), fileO);
					fputs("\n", fileO);
				}

				if (!structure.con1.empty()) {
					fputs(".names ", fileO);
					fputs(structure.con1.c_str(), fileO);
					fputs("\n1\n", fileO);
				}
			}
			lineCount++;
		}

		readLine(fileI, line);
	}
	fclose(fileO);
}

void writeDotForDecomp(Decomp *decomp, char* fileName) {
	FILE *file = fopen(fileName, "w");

	DdNode **outputs = new DdNode*[1];
	outputs[0] = Cudd_BddToAdd(decomp -> manag, decomp -> function);
	Cudd_Ref(outputs[0]);

	Cudd_DumpDot(decomp -> manag, 1, outputs, NULL, NULL, file);
	fclose(file);
}
